﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

// Networking
using Google.Protobuf;
using TFG.DataLayer;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace Pixotope
{
    static class DataLayer
    {

        // FRONTIER DEVICES
        
        public static List<PixotopeDevice> PixotopeDevices;


        // CONNECTION

        private static UdpClient UdpClient;
        public static SenderInfo SenderData;

        private static bool HandshakeSuccessful = false;

        private static DateTime PingStarted;
        public static float PingLatency;


        // CONNECTION SETTINGS

        private static IPAddress DataLayerIP = null;
        private static int DataLayerPort = 5005;


        // MESSAGE SEND QUEUE

        private static Queue<UnionRequest> RequestQueue = new Queue<UnionRequest>();
        private static int RequestQueueMaximumLength = 20;
        private static int SendDelay = 50;


        // CONNECTION SETUP

        private static void ClearConnection()
        {
            // Clear handshake flag
            HandshakeSuccessful = false;

            // Clear Pixotope device list
            PixotopeDevices = new List<PixotopeDevice>();
           
            // Clear UDP client
            if (UdpClient != null) UdpClient.Close();
            UdpClient = null;
        }

        public static void Disconnect()
        {
            // Clear connection
            ClearConnection();
        }

        public static void Connect(string receiverIP, string receiverPort, string senderIP)
        {
            // Clear connection
            ClearConnection();

            // Sanitize input
            IPAddress tempIP = null;
            IPAddress.TryParse(receiverIP, out tempIP);
            int tempPort = Int32.Parse(receiverPort);
            if (tempIP != null)
            {
                // Assign input values
                DataLayerIP = tempIP;
                DataLayerPort = tempPort;

                // Construct sender identity
                SenderData = new SenderInfo();
                SenderData.ID = "Flowgic";
                SenderData.IP = senderIP;

                // Construct UDP client object
                if (UdpClient == null)
                {
                    UdpClient = new UdpClient(0);
                }
                SenderData.Port = (uint)((IPEndPoint)UdpClient.Client.LocalEndPoint).Port;

                // Begin UDP Listener
                UdpListener();

                // Begin UDP Sender
                UdpSender();

                // Begin server connection communication chain
                SendHandshakeRequest();
            }
        }

        private static void OnConnectionSuccessfull()
        {

        }


        // FRONTIER DEVICES

        private static void EnlistPixotopeDevices(List<string> PixotopeDeviceNames)
        {
            // Non-destructive clearing
            for (int i = PixotopeDevices.Count - 1; i >= 0; i--)
            {
                if (!PixotopeDeviceNames.Contains(PixotopeDevices[i].ID))
                {
                    PixotopeDevices.RemoveAt(i);
                }
            }

            // Instantiate PixotopeDevice object for each name
            foreach (string name in PixotopeDeviceNames)
            {
                // Check if the name is already present
                bool isNamePresent = false;
                foreach (PixotopeDevice device in PixotopeDevices)
                {
                    if (device.ID == name)
                    {
                        isNamePresent = true;
                        break;
                    }
                }

                // Create new PixotopeDevice if not
                if (!isNamePresent)
                {
                    PixotopeDevice newPixotopeDevice = new PixotopeDevice(name);
                    PixotopeDevices.Add(newPixotopeDevice);
                }
            }
        }
        
        public static void ChangePixotopeDeviceSetTargetFlag(string id, bool value)
        {
            // Iterate over delcared Pixotope devices
            foreach (PixotopeDevice device in PixotopeDevices) if (id.Equals(device.ID)) device.IsSetTarget = value;
        }

        public static List<PixotopeDevice> GetSetTargets()
        {
            // Declare container
            List<PixotopeDevice> devices = new List<PixotopeDevice>();

            // Iterate over delcared Pixotope devices
            if (PixotopeDevices != null)
            {
                foreach (PixotopeDevice device in PixotopeDevices)
                {
                    if (device.IsSetTarget) devices.Add(device);
                }
            }

            // Return container
            return devices;
        }

        public static List<string> GetSetTargetIDs()
        {
            List<string> devices = new List<string>();
            foreach (PixotopeDevice device in GetSetTargets()) devices.Add(device.ID);
            return devices;
        }
                

        // SENDING

        private static void SendHandshakeRequest()
        {
            // Construct request
            UnionRequest UnionRequest = new UnionRequest();
            UnionRequest.HSReq = new HandshakeRequest();
            UnionRequest.HSReq.SenderData = SenderData;
            UnionRequest.HSReq.SenderData.TimeStamp = GetEpochTimeStamp();

            // Send command
            SendUnionRequest(UnionRequest);
        }

        private static void SendSearchForPixotopeDevices()
        {
            // Construct request
            UnionRequest UnionRequest = new UnionRequest();
            UnionRequest.SearchReq = new ProviderSearchRequest();
            UnionRequest.SearchReq.SenderData = SenderData;
            UnionRequest.SearchReq.SearchString = "_Editor";
            UnionRequest.SearchReq.ReturnAddress = false;

            // Send command
            SendUnionRequest(UnionRequest);
        }

        public static void SendPingRequest(bool requestReply = true)
        {
            // Set time record for latency measurement
            PingStarted = DateTime.Now;

            // Construct request
            UnionRequest UnionRequest = new UnionRequest();
            UnionRequest.PingReq = new PingRequest();
            UnionRequest.PingReq.SenderData = SenderData;
            UnionRequest.PingReq.RequestReply = requestReply;
            UnionRequest.PingReq.SenderData.TimeStamp = GetEpochTimeStamp();

            // Send command
            SendUnionRequest(UnionRequest);
        }

        public static void SendDataRequest(DataPacket Data)
        {
            // Construct request
            UnionRequest UnionRequest = new UnionRequest();
            UnionRequest.DataReq = new SetDataRequest();
            UnionRequest.DataReq.SenderData = SenderData;
            UnionRequest.DataReq.Data.Add(Data);
            UnionRequest.DataReq.SenderData.TimeStamp = GetEpochTimeStamp();
            UnionRequest.DataReq.Data[0].TimeStamp = GetEpochTimeStamp();

            // Send command
            SendUnionRequest(UnionRequest);
        }
        
        public static void SendCommandRequest(List<string> receivers, string command, List<string> arguments)
        {
            // Check if there are any target receivers
            if (receivers != null && receivers.Count > 0)
            {
                // Construct request
                UnionRequest UnionRequest = new UnionRequest();
                UnionRequest.CmndReq = new SetCommandRequest();
                UnionRequest.CmndReq.SenderData = SenderData;
                foreach (string receiver in receivers) UnionRequest.CmndReq.ReceiverID.Add(receiver);
                // Construct command packet
                UnionRequest.CmndReq.Command = new CommandPacket();
                UnionRequest.CmndReq.Command.TimeStamp = GetEpochTimeStamp();
                foreach (string receiver in receivers) UnionRequest.CmndReq.Command.ReceiverID.Add(receiver);
                UnionRequest.CmndReq.Command.Command = command;
                if (arguments != null) foreach (string argument in arguments) UnionRequest.CmndReq.Command.Arguments.Add(argument);
                // Send command
                SendUnionRequest(UnionRequest);
            }
        }


        // SEND & RECEIVE

        public static void SendUnionRequest(UnionRequest UnionRequest)
        {
            // Enqueue
            RequestQueue.Enqueue(UnionRequest);

            // Remove overflow
            lock (RequestQueue)
            {
                while (RequestQueue.Count() > RequestQueueMaximumLength) RequestQueue.Dequeue();
            }
        }

        private static void UdpSender()
        {
            Task.Run(async () =>
            {
                while (true)
                {
                    // Send if there is a message queued
                    if (RequestQueue.Count() > 0)
                    {
                        lock (RequestQueue)
                        {
                            // Dequeue
                            UnionRequest request = RequestQueue.Dequeue();

                            // Check if action request was formed
                            if (request != null)
                            {
                                // Convert the object to binary using Google Protobuf library
                                Byte[] Bytes = request.ToByteArray();

                                // Send
                                {
                                    try
                                    {
                                        // Create IP endPoint
                                        IPEndPoint endPoint = new IPEndPoint(DataLayerIP, DataLayerPort);

                                        // Send the byte array to the server
                                        UdpClient.Send(Bytes, Bytes.Length, endPoint);
                                    }
                                    catch { }
                                }
                            }
                        }
                    }

                    // Wait
                    Thread.Sleep(SendDelay);
                }
            });
        }

        private static void ReceiveUnionRequest(UnionRequest UnionRequest)
        {
            // Data Response
            if (UnionRequest.DataReq != null && UnionRequest.DataReq.Data != null && UnionRequest.DataReq.Data.Count > 0)
            {                  
                // Reflection Error
                if (UnionRequest.DataReq.Data[0].Type.Equals("ReflectionError"))
                {
                    // Read error messages
                    List<string> messages = UnionRequest.DataReq.Data[0].Data.Split(',').ToList();
                }

                // Function Return
                else if (UnionRequest.DataReq.Data[0].Type.Equals("FunctionReturn"))
                {
                    // Interpret received data
                    string[] Data = UnionRequest.DataReq.Data[0].Data.Split('|');
                    if (Data.Length >= 3)
                    {
                        // Deserialize
                        string ActorName = Data[0];
                        string FunctionName = Data[1];
                        List<string> ReturnedValues = new List<string>();
                        for (int i = 2; i < Data.Length; i++) ReturnedValues.Add(Data[i]);
                    }
                }
            }

            // Handshake Response
            if (UnionRequest.HSRes != null)
            {
                // Set flag
                HandshakeSuccessful = true;

                // Request Pixotope names
                SendSearchForPixotopeDevices();

                // Connection success
                OnConnectionSuccessfull();
            }

            // Provider Search Response
            if (UnionRequest.SearchRes != null)
            {
                // Pixotopes
                if (UnionRequest.SearchRes.SearchString.Equals("_Editor"))
                {
                    // Load pixotope device list
                    List<ProviderSearchResult> SearchResults = UnionRequest.SearchRes.SearchResults.ToList();
                    List<string> PixotopeDeviceNames = new List<string>();
                    foreach (ProviderSearchResult searchResult in SearchResults) PixotopeDeviceNames.Add(searchResult.ID);
                    EnlistPixotopeDevices(PixotopeDeviceNames);
                }
            }

            // Ping Response
            if (UnionRequest.PingReq != null)
            {
                if (UnionRequest.PingReq.RequestReply)
                {
                    // Respond to heart beat ping
                    SendPingRequest(false);
                }
                else
                {
                    // Evaluate response to ping
                    PingLatency = (float)DateTime.Now.Subtract(PingStarted).TotalSeconds;
                }
            }

            // Command Request
            if (UnionRequest.CmndReq != null)
            {
                // Check if valid
                if (UnionRequest.CmndReq.Command != null && UnionRequest.CmndReq.SenderData != null)
                {
                    // Get commands
                    //string senderID = UnionRequest.CmndReq.SenderData.ID;
                    //List<string> commands = UnionRequest.CmndReq.Command.CommandList.ToList();
                }
            }
        }

        private static void UdpListener()
        {
            Task.Run(async () =>
            {
                // Receive
                while (true)
                {
                    var receivedResults = await UdpClient.ReceiveAsync();

                    UnionRequest UnionRequest = new UnionRequest();
                    UnionRequest.MergeFrom(receivedResults.Buffer);
                    ReceiveUnionRequest(UnionRequest);
                }
            });
        }


        // FRONTIER MESSAGES

        public static void CallFunction(string ActorName, string FunctionName, List<string> Arguments)
        {
            // Check if data connection is established
            if (DataLayer.IsConnected())
            {
                // Set arguments
                List<string> arguments = new List<string>();
                arguments.Add(ActorName);
                arguments.Add(FunctionName);
                arguments.AddRange(Arguments);
                // Send command
                SendCommandRequest(GetSetTargetIDs(), "CallFunction", arguments);
            }
        }


        // UTILS

        public static ulong GetEpochTimeStamp()
        {
            TimeSpan t = DateTime.UtcNow - new DateTime(1970, 1, 1);
            return (ulong)(t.Ticks / 10000.0f);
        }

        public static bool IsPixotopeReady()
        {
            return IsConnected() && PixotopeDevices != null;
        }
        
        public static bool IsConnected()
        {
            return HandshakeSuccessful;
        }
    }
}
